import socket, time, re, json, threading, select
from struct import pack

import requests

from .Abstract import AbstractDanMuClient

class _socket(socket.socket):
    def push(self, data, t=b'\x01'):
        data = t + pack('>I', len(data))[1:] + data
        self.sendall(data)
    def pull(self):
        try: # for socket.settimeout
            return self.recv(9999)
        except Exception as e:
            return ''

class HuoMaoDanMuClient(AbstractDanMuClient):
    def _get_live_status(self):
        c = requests.get(self.url).content
        r = re.search('getFlash\("(.*?)","(.*?)"\);', c)
        data = {
            'cid': r.group(1),
            'cdns': '1',
            'streamtype': 'live',
            'VideoIDS': r.group(2), }
        j = requests.post('http://www.huomao.com/swf/live_data', data=data).json()
        return j['roomStatus'] == '1'
    def _prepare_env(self):
        c = requests.get(self.url).content
        r = re.search('getFlash\("(.*?)","(.*?)"\);', c)
        roomId = r.group(1)
        url = 'http://chat.huomao.com/chat/getToken'
        params = {
            'callback': 'jQuery171032695039477104815_1477741089191',
            'cid': roomId,
            '_': int(time.time() * 100), }
        headers = { 'User-Agent': 'User-Agent: Mozilla/5.0 (Windows NT 6.3; Win64; x64)'\
            'AppleWebKit/537.36 (KHTML, like Gecko) Chrome/54.0.2840.71 Safari/537.36', }
        c = requests.get(url, params=params, headers=headers).content
        r = re.search('jQuery[^(]*?\((.*?)\)$', c)
        j = json.loads(r.group(1))['data']

        s = _socket()
        s.connect((j['host'], int(j['port'])))
        data = {
            'user': None,
            'sys': {
                'version': '0.1.6b',
                'pomelo_version': '0.7.x',
                'type': 'pomelo-flash-tcp', }, }
        data = json.dumps(data, separators=(',', ':'))
        s.push(data)
        s.pull()

        s.sendall(b'\x02\x00\x00\x00')

        data = {
            'channelId': roomId,
            'log': True,
            'userId': '', }
        data = '\x00\x01\x20gate.gateHandler.lookupConnector'\
            + json.dumps(data, separators=(',', ':'))
        s.push(data, b'\x04')
        r = s.pull()[6:]
        danmuPort = json.loads(r)

        return (danmuPort['host'], int(danmuPort['port'])),\
                {'roomId': roomId, 'token': j['token'], 'userId': j['uid']}
    def _init_socket(self, danmu, roomInfo):
        self.danmuSocket = _socket()
        self.danmuSocket.connect(danmu)
        self.danmuSocket.settimeout(3)

        data = {
            'user': None,
            'sys': {
                'version': '0.1.6b',
                'pomelo_version': '0.7.x',
                'type': 'pomelo-flash-tcp', }, }
        data = json.dumps(data, separators=(',', ':'))
        self.danmuSocket.push(data)
        self.danmuSocket.pull()

        self.danmuSocket.sendall(b'\x02\x00\x00\x00')
        self.danmuSocket.pull()

        data = {
            'channelId': roomInfo['roomId'],
            'token': roomInfo['token'],
            'userId': roomInfo['userId'], }
        data = json.dumps(data, separators=(',', ':'))
        data = '\x00\x02\x20' + 'connector.connectorHandler.login' + data
        self.danmuSocket.push(data, b'\x04')
    def _create_thread_fn(self, roomInfo):
        def keep_alive(self):
            time.sleep(30)
            self.danmuSocket.sendall(b'\x03\x00\x00\x00')
        def get_danmu(self):
            if not select.select([self.danmuSocket], [], [], 1)[0]: return
            content = self.danmuSocket.pull()
            for msg in re.findall(b'\x04\x00.*?({"[^\x04]*})', content):
                try:
                    msg = json.loads(msg.decode('utf8', 'ignore'))
                    if 'msg_content' not in msg or 'msg_type' not in msg: continue
                    msg['NickName'] = msg['msg_content']['username']
                    msg['Content']  = msg['msg_content'].get('content') or \
                        msg['msg_content'].get('amount')
                    msg['MsgType']  = {'msg': 'danmu', 'beans': 'gift',
                        'welcome': 'enter'}.get(msg.get('msg_type'), 'other')
                except Exception as e:
                    pass
                else:
                    self.danmuWaitTime = time.time() + self.maxNoDanMuWait
                    self.msgPipe.append(msg)
        return get_danmu, keep_alive # danmu, heart



import socket, json, re, select, time, random
from struct import pack

import requests

from .Abstract import AbstractDanMuClient

class _socket(socket.socket):
    def push(self, data, type = 7):
        data = (pack('>i', len(data) + 16) + b'\x00\x10\x00\x01' +
            pack('>i', type) + pack('>i', 1) + data)
        self.sendall(data)
    def pull(self):
        try: # for socket.settimeout
            return self.recv(9999)
        except Exception as e:
            return ''

class BilibiliDanMuClient(AbstractDanMuClient):
    def _get_live_status(self):
        url = ('http://live.bilibili.com/'
            + self.url.split('/')[-1] or self.url.split('/')[-2])
        self.roomId = re.findall(b'var ROOMID = (\d+);', requests.get(url).content)[0].decode('ascii')
        r = requests.get('http://live.bilibili.com/api/player?id=cid:' + self.roomId)
        self.serverUrl = re.findall(b'<server>(.*?)</server>', r.content)[0].decode('ascii')
        return re.findall(b'<state>(.*?)</state>', r.content)[0] == b'LIVE'
    def _prepare_env(self):
        return (self.serverUrl, 788), {}
    def _init_socket(self, danmu, roomInfo):
        self.danmuSocket = _socket()
        self.danmuSocket.connect(danmu)
        self.danmuSocket.settimeout(3)
        self.danmuSocket.push(data = json.dumps({
            'roomid': int(self.roomId),
            'uid': int(1e14 + 2e14 * random.random()),
            }, separators=(',', ':')).encode('ascii'))
    def _create_thread_fn(self, roomInfo):
        def keep_alive(self):
            self.danmuSocket.push(b'', 2)
            time.sleep(30)
        def get_danmu(self):
            if not select.select([self.danmuSocket], [], [], 1)[0]: return
            content = self.danmuSocket.pull()
            for msg in re.findall(b'\x00({[^\x00]*})', content):
                try:
                    msg = json.loads(msg.decode('utf8', 'ignore'))
                    msg['NickName'] = (msg.get('info', ['','',['', '']])[2][1]
                        or msg.get('data', {}).get('uname', ''))
                    msg['Content']  = msg.get('info', ['', ''])[1]
                    msg['MsgType']  = {'SEND_GIFT': 'gift', 'DANMU_MSG': 'danmu',
                        'WELCOME': 'enter'}.get(msg.get('cmd'), 'other')
                except Exception as e:
                    pass
                else:
                    self.danmuWaitTime = time.time() + self.maxNoDanMuWait
                    self.msgPipe.append(msg)
        return get_danmu, keep_alive # danmu, heart



import time, sys, re, json
import socket, select
from struct import pack

import requests

from .Abstract import AbstractDanMuClient

class _socket(socket.socket):
    def communicate(self, data):
        self.push(data)
        return self.pull()
    def push(self, data):
        self.sendall(data)
    def pull(self):
        try: # for socket.settimeout
            return self.recv(9999)
        except:
            return ''

class PandaDanMuClient(AbstractDanMuClient):
    def _get_live_status(self):
        params = {
            'roomid': (self.url.split('/')[-1] or
                self.url.split('/')[-2]),
            'pub_key': '',
            '_': int(time.time()), }
        j = requests.get('http://www.panda.tv/api_room', params).json()['data']
        return j['videoinfo']['status'] == '2'
    def _prepare_env(self):
        roomId = self.url.split('/')[-1] or self.url.split('/')[-2]
        url = 'http://www.panda.tv/ajax_chatroom?roomid=%s&_=%s'%(roomId, str(int(time.time())))
        roomInfo = requests.get(url).json()
        url = 'http://api.homer.panda.tv/chatroom/getinfo'
        params = {
            'rid': roomInfo['data']['rid'],
            'roomid': roomId,
            'retry': 0,
            'sign': roomInfo['data']['sign'], 
            'ts': roomInfo['data']['ts'],
            '_': int(time.time()), }
        serverInfo = requests.get(url, params).json()['data']
        serverAddress = serverInfo['chat_addr_list'][0].split(':')
        return (serverAddress[0], int(serverAddress[1])), serverInfo
    def _init_socket(self, danmu, roomInfo):
        data = [
            ('u', '%s@%s'%(roomInfo['rid'], roomInfo['appid'])),
            ('k', 1),
            ('t', 300),
            ('ts', roomInfo['ts']),
            ('sign', roomInfo['sign']),
            ('authtype', roomInfo['authType']) ]
        data = '\n'.join('%s:%s'%(k, v) for k, v in data)
        data = (b'\x00\x06\x00\x02\x00' + pack('B', len(data)) +
            data.encode('utf8') + b'\x00\x06\x00\x00')
        self.danmuSocket = _socket(socket.AF_INET, socket.SOCK_STREAM)
        self.danmuSocket.settimeout(3)
        self.danmuSocket.connect(danmu)
        self.danmuSocket.push(data)
    def _create_thread_fn(self, roomInfo):
        def get_danmu(self):
            if not select.select([self.danmuSocket], [], [], 1)[0]: return
            content = self.danmuSocket.pull()
            for msg in re.findall(b'({"type":.*?}})', content):
                try:
                    msg = json.loads(msg.decode('utf8', 'ignore'))
                    msg['NickName'] = msg.get('data', {}).get('from', {}
                        ).get('nickName', '')
                    msg['Content']  = msg.get('data', {}).get('content', '')
                    msg['MsgType']  = {'1': 'danmu', '206': 'gift'
                        }.get(msg['type'], 'other')
                except:
                    pass
                else:
                    self.danmuWaitTime = time.time() + self.maxNoDanMuWait
                    self.msgPipe.append(msg)
        def heart_beat(self):
            self.danmuSocket.push(b'\x00\x06\x00\x06')
            time.sleep(60)
        return get_danmu, heart_beat



import os, logging

if not os.path.exists('config'): os.mkdir('config')
log = logging.getLogger('danmu')
log.setLevel(logging.DEBUG)
fileHandler = logging.FileHandler(os.path.join('config', 'run.log'), encoding = 'utf8')
fileHandler.setLevel(logging.DEBUG)
formatter = logging.Formatter('%(asctime)-17s <%(message)s> %(levelname)s %(filename)s[%(lineno)d]',
    datefmt='%Y%m%d %H:%M:%S')
fileHandler.setFormatter(formatter)
log.addHandler(fileHandler)

if __name__ == '__main__':
    log.debug('This is debug')
    log.info('This is info')



import time, socket, select, sys, re, json
from struct import pack, unpack

import requests

from .Abstract import AbstractDanMuClient

class _socket(socket.socket):
    def communicate(self, data):
        self.push(data)
        return self.pull()
    def push(self, data):
        self.sendall(data)
    def pull(self):
        try: # for socket.settimeout
            return self.recv(9999)
        except:
            return ''

class QuanMinDanMuClient(AbstractDanMuClient):
    def _get_live_status(self):
        url = '%s/%s/info.json?t=%s' % ('http://www.quanmin.tv/json/rooms',
            self.url.split('/')[-1] or self.url.split('/')[-2], int(time.time() / 50))
        return requests.get(url).json()['play_status']
    def _prepare_env(self):
        r = requests.get('http://www.quanmin.tv/site/route?time='
                + str(int(time.time()))).content
        danmuIp = '.'.join([str(i ^ 172) for i in unpack('>iiii', r[:16])])
        roomId = self.url.split('/')[-1] or self.url.split('/')[-2]
        if roomId.isdigit():
            roomInfo = {'uid': int(roomId), }
        else:
            url = '%s/%s/info.json?t=%s'%('http://www.quanmin.tv/json/rooms', 
                roomId, int(time.time() / 50))
            roomInfo = requests.get(url).json()
        return (danmuIp, 9098), roomInfo # danmu, roomInfo, success
    def _init_socket(self, danmu, roomInfo):
        self.danmuSocket = _socket()
        self.danmuSocket.settimeout(3)
        roomId = roomInfo['uid']
        data = ('{\n' +
            '   "os" : 135,\n'
            '   "pid" : 10003,\n'
            '   "rid" : "%s",\n'
            '   "timestamp" : 78,\n'
            '   "ver" : 147\n}')%roomId
        data = pack('>i', len(data)) + data.encode('ascii') + b'\x0a'
        self.danmuSocket.connect(danmu)
        self.danmuSocket.push(data)
    def _create_thread_fn(self, roomInfo):
        def get_danmu(self):
            if not select.select([self.danmuSocket], [], [], 1)[0]: return
            content = self.danmuSocket.pull()
            for msg in re.findall(b'\x00\x00.*?({[^\x00]*)', content):
                try:
                    msg = json.loads(json.loads(msg.decode('ascii'))['chat']['json'])
                    msg['NickName'] = msg.get('user', {}).get('nick', '')
                    msg['Content']  = msg.get('text', '')
                    if msg.get('type') in (1,2,3,4,5):
                        msg['MsgType'] = 'gift'
                    elif msg.get('type') == 0:
                        msg['MsgType'] = 'danmu'
                    else:
                        msg['MsgType'] = 'other'
                except Exception as e:
                    pass
                else:
                    self.danmuWaitTime = time.time() + self.maxNoDanMuWait
                    self.msgPipe.append(msg)
        def heart_beat(self):
            time.sleep(3)
        return get_danmu, heart_beat # danmu, heart



import re, threading, time, traceback

from .DouYu    import DouYuDanMuClient
from .Panda    import PandaDanMuClient
from .ZhanQi   import ZhanQiDanMuClient
from .QuanMin  import QuanMinDanMuClient
from .Bilibili import BilibiliDanMuClient
from .HuoMao   import HuoMaoDanMuClient

__version__ = '1.0.4'
__all__     = ['DanMuClient']

class DanMuClient(object):
    def __init__(self, url):
        self.__url          = ''
        self.__baseClient   = None
        self.__client       = None
        self.__functionDict = {'default': lambda x: 0}
        self.__isRunning    = False
        if 'http://' == url[:7]:
            self.__url = url
        else:
            self.__url = 'http://' + url
        for u, bc in {'panda.tv'    : PandaDanMuClient,
                'douyu.com'         : DouYuDanMuClient,
                'quanmin.tv'        : QuanMinDanMuClient,
                'zhanqi.tv'         : ZhanQiDanMuClient,
                'live.bilibili.com' : BilibiliDanMuClient,
                'huomao.com'        : HuoMaoDanMuClient, }.items() :
            if re.match(r'^(?:http://)?.*?%s/(.+?)$' % u, url):
                self.__baseClient = bc; break
    def __register(self, fn, msgType):
        if fn is None:
            if msgType == 'default':
                self.__functionDict['default'] = lambda x: 0
            elif self.__functionDict.get(msgType):
                del self.__functionDict[msgType]
        else:
            self.__functionDict[msgType] = fn
    def isValid(self):
        return self.__baseClient is not None
    def default(self, fn):
        self.__register(fn, 'default')
        return fn
    def danmu(self, fn):
        self.__register(fn, 'danmu')
        return fn
    def gift(self, fn):
        self.__register(fn, 'gift')
        return fn
    def other(self, fn):
        self.__register(fn, 'other')
        return fn
    def start(self, blockThread = False, pauseTime = .1):
        if not self.__baseClient or self.__isRunning: return False
        self.__client = self.__baseClient(self.__url)
        self.__isRunning = True
        receiveThread = threading.Thread(target=self.__client.start)
        receiveThread.setDaemon(True)
        receiveThread.start()
        def _start():
            while self.__isRunning:
                if self.__client.msgPipe:
                    msg = self.__client.msgPipe.pop()
                    fn = self.__functionDict.get(msg['MsgType'],
                        self.__functionDict['default'])
                    try:
                        fn(msg)
                    except:
                        traceback.print_exc()
                else:
                    time.sleep(pauseTime)
        if blockThread:
            try:
                _start()
            except KeyboardInterrupt:
                print('Bye~')
        else:
            danmuThread = threading.Thread(target = _start)
            danmuThread.setDaemon(True)
            danmuThread.start()
        return True
    def stop(self):
        self.__isRunning = False
        if self.__client: self.__client.deprecated = True



import abc, threading, time, logging

logger = logging.getLogger('danmu')

# This client will auto-reload if exception is raised inside and write a log
# it is deprecated once used
# log of main start and thread is recorded
# If you want to close it outside, just set the deprecated flag to True
# Inside reload is controlled by self.live flag
# danmuWaitTime is wrapped in danmuThread
# this client may cause unclosed thread because of thread is blocked, but it's not a big problem
class AbstractDanMuClient(object):
    __metaclass__ = abc.ABCMeta

    def __init__(self, url, maxNoDanMuWait = 180, anchorStatusRescanTime = 30):
        self.url = url
        self.maxNoDanMuWait = maxNoDanMuWait
        self.anchorStatusRescanTime = anchorStatusRescanTime
        self.deprecated = False # this is an outer live flag
        self.live = False # this is an inner live flag
        self.danmuSocket = None
        self.danmuThread, self.heartThread = None, None
        self.msgPipe = []
        self.danmuWaitTime = -1
    def start(self):
        while not self.deprecated:
            try:
                while not self.deprecated:
                    if self._get_live_status(): break
                    time.sleep(self.anchorStatusRescanTime)
                else:
                    break
                danmuSocketInfo, roomInfo = self._prepare_env()
                if self.danmuSocket: self.danmuSocket.close()
                self.danmuWaitTime = -1
                self._init_socket(danmuSocketInfo, roomInfo)
                danmuThreadFn, heartThreadFn = self._create_thread_fn(roomInfo)
                self._wrap_thread(danmuThreadFn, heartThreadFn)
                self._start_receive()
            except Exception as e:
                logger.debug(str(e.args))
                time.sleep(5)
            else:
                break
    def _socket_timeout(self, fn):
    # if socket went wrong, reload the whole client
        def __socket_timeout(*args, **kwargs):
            try:
                fn(*args, **kwargs)
            except Exception as e:
                logger.debug(str(e.args))
                if not self.live: return
                self.live = False
                # In case thread is blocked and can't stop, set a max wait time
                waitEndTime = time.time() + 20
                while self.thread_alive() and time.time() < waitEndTime:
                    time.sleep(1)
                self.start()
        return __socket_timeout
    def _wrap_thread(self, danmuThreadFn, heartThreadFn):
        @self._socket_timeout
        def heart_beat(self):
            while self.live and not self.deprecated:
                heartThreadFn(self)
        @self._socket_timeout
        def get_danmu(self):
            while self.live and not self.deprecated:
                if self.danmuWaitTime != -1 and self.danmuWaitTime < time.time():
                    raise Exception('No danmu received in %ss'%self.maxNoDanMuWait)
                danmuThreadFn(self)
        self.heartThread = threading.Thread(target = heart_beat, args = (self,))
        self.heartThread.setDaemon(True)
        self.danmuThread = threading.Thread(target = get_danmu, args = (self,))
        self.danmuThread.setDaemon(True)
    def _start_receive(self):
        self.live = True
        self.danmuThread.start()
        self.heartThread.start()
        self.danmuWaitTime = time.time() + 20
    def thread_alive(self):
        if self.danmuSocket is None or not self.danmuThread.isAlive():
            return False
        else:
            return True
    @abc.abstractmethod
    def _get_live_status(self):
        return False # return whether anchor is on live
    @abc.abstractmethod
    def _prepare_env(self):
        return ('0.0.0.0', 80), {} # danmu, roomInfo
    @abc.abstractmethod
    def _init_socket(self, danmu, roomInfo):
        pass
    # method shouldn't include the main while loop
    # danmu method should reload self.danmuWaitTime
    @abc.abstractmethod
    def _create_thread_fn(self, roomInfo):
        return None, None # danmu, heart

class DanMuException(Exception):
    def __init__(self, message, *args, **kwargs):
        Exception.__init__(self)
        self.message = message
        self.args = args
    def __str__(self):
        return self.message



import socket, json, re, select, time
from struct import pack

import requests

from .Abstract import AbstractDanMuClient

class _socket(socket.socket):
    def communicate(self, data):
        self.push(data)
        return self.pull()
    def push(self, data):
        s = pack('i', 9 + len(data)) * 2
        s += b'\xb1\x02\x00\x00' # 689
        s += data.encode('ascii') + b'\x00'
        self.sendall(s)
    def pull(self):
        try: # for socket.settimeout
            return self.recv(9999)
        except Exception as e:
            return ''

class DouYuDanMuClient(AbstractDanMuClient):
    def _get_live_status(self):
        url = 'http://open.douyucdn.cn/api/RoomApi/room/%s' % (
            self.url.split('/')[-1] or self.url.split('/')[-2])
        j = requests.get(url).json()
        if j.get('error') != 0 or j['data'].get('room_status') != '1': return False
        self.roomId = j['data']['room_id']
        return True
    def _prepare_env(self):
        return ('openbarrage.douyutv.com', 8601), {'room_id': self.roomId}
    def _init_socket(self, danmu, roomInfo):
        self.danmuSocket = _socket()
        self.danmuSocket.connect(danmu)
        self.danmuSocket.settimeout(3)
        self.danmuSocket.communicate('type@=loginreq/roomid@=%s/'%roomInfo['room_id'])
        self.danmuSocket.push('type@=joingroup/rid@=%s/gid@=-9999/'%roomInfo['room_id'])
    def _create_thread_fn(self, roomInfo):
        def keep_alive(self):
            self.danmuSocket.push('type@=keeplive/tick@=%s/'%int(time.time()))
            time.sleep(30)
        def get_danmu(self):
            if not select.select([self.danmuSocket], [], [], 1)[0]: return
            content = self.danmuSocket.pull()
            for msg in re.findall(b'(type@=.*?)\x00', content):
                try:
                    msg = msg.replace(b'@=', b'":"').replace(b'/', b'","')
                    msg = msg.replace(b'@A', b'@').replace(b'@S', b'/')
                    msg = json.loads((b'{"' + msg[:-2] + b'}').decode('utf8', 'ignore'))
                    msg['NickName'] = msg.get('nn', '')
                    msg['Content']  = msg.get('txt', '')
                    msg['MsgType']  = {'dgb': 'gift', 'chatmsg': 'danmu',
                        'uenter': 'enter'}.get(msg['type'], 'other')
                except Exception as e:
                    pass
                else:
                    self.danmuWaitTime = time.time() + self.maxNoDanMuWait
                    self.msgPipe.append(msg)
        return get_danmu, keep_alive # danmu, heart



import sys, json, time, re, base64
import socket, threading, select
from struct import pack

import requests

from .Abstract import AbstractDanMuClient

class ZhanQiDanMuClient(AbstractDanMuClient):
    def _get_live_status(self):
        r = requests.get('http://www.zhanqi.tv/' +
            self.url.split('/')[-1] or self.url.split('/')[-2])
        if r.url == 'http://www.zhanqi.tv/': return False
        rawJson = re.findall('oRoom = (.*);[\s\S]*?window.bClose', r.text)
        if not rawJson: rawJson = re.findall('aVideos = (.*);[\s\S]*?oPageConfig.oUrl', r.text)
        self.roomInfo = json.loads(rawJson[0])
        if isinstance(self.roomInfo, list): self.roomInfo = self.roomInfo[0]
        return self.roomInfo['status'] == '4'
    def _prepare_env(self):
        serverAddress = json.loads(base64.b64decode(
            self.roomInfo['flashvars']['Servers']).decode('ascii'))['list'][0]
        serverAddress = (serverAddress['ip'], serverAddress['port'])
        url = '%s/api/public/room.viewer' % 'http://www.zhanqi.tv'
        params = {
            'uid': self.roomInfo['uid'],
            '_t': int(time.time() / 60), }
        roomInfo = requests.get(url, params).json()
        roomInfo['id'] = int(self.roomInfo['id'])
        return serverAddress, roomInfo # danmu, roomInfo
    def _init_socket(self, danmu, roomInfo):
        self.danmuSocket = socket.socket()
        self.danmuSocket.settimeout(3)
        data = {
            'nickname'     : '',
            'roomid'       : int(roomInfo['id']),
            'gid'          : roomInfo['data']['gid'],
            'sid'          : roomInfo['data']['sid'],
            'ssid'         : roomInfo['data']['sid'],
            'timestamp'    : roomInfo['data']['timestamp'],
            'cmdid'        : 'loginreq',
            'develop_date' : '2015-06-07',
            'fhost'        : 'zhanqi.tool',
            'fx'           : 0,
            't'            : 0,
            'thirdacount'  : '',
            'uid'          : 0,
            'ver'          : 2,
            'vod'          : 0, }
        data = json.dumps(data, separators = (',',':'))
        self.danmuSocket.connect(danmu)
        self.danmuSocket.sendall(b'\xbb\xcc' + b'\x00'*4 +
            pack('i', len(data)) + b'\x10\x27' + data.encode('ascii'))
    def _create_thread_fn(self, roomInfo):
        def get_danmu(self):
            if not select.select([self.danmuSocket], [], [], 1)[0]: return
            content = self.danmuSocket.recv(999)
            for msg in re.findall(b'\x10\x27({[^\x00]*})\x0a', content):
                try:
                    msg = json.loads(msg.decode('utf8', 'ignore'))
                    msg['NickName'] = (msg.get('fromname', '') or
                        msg.get('data', {}).get('nickname', ''))
                    msg['Content']  = msg.get('content', '')
                    if 'chatm' in msg.get('cmdid', ''):
                        msg['MsgType'] = 'danmu'
                    elif 'Gift' in msg.get('cmdid', ''):
                        msg['MsgType'] = 'gift'
                    else:
                        msg['MsgType'] = 'other'
                except Exception as e:
                    pass
                else:
                    self.danmuWaitTime = time.time() + self.maxNoDanMuWait
                    self.msgPipe.append(msg)
        def heart_beat(self):
            time.sleep(3)
            self.danmuSocket.sendall(b'\xbb\xcc' + b'\x00'*8 + b'\x59\x27')
        return get_danmu, heart_beat # danmu, heart
